/*-
 * Copyright (c) 2008, Alexandre P. Francisco <aplf@ist.utl.pt>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _RAND_CMP
#include <sys/time.h>
#endif /* _RAND_CMP */

#ifdef _TIMER
#include <time.h>
#endif /* _TIMER */

#include <assert.h>

#include "gcfnd.h"
#include "unfnd.h"
#include "pqueue.h"

/*
 * Edge access macros.
 */
#define e_nxt(t,x) (((x)->u == (t)) ? (x)->u_nxt : (x)->v_nxt)
#define e_adj(t,x) (((x)->u == (t)) ? (x)->v : (x)->u)

/*
 * Given the adjacency list for each vertex and the list of edges, the
 * edge 'e' is added to the adjacency list of vertex 'x', if possible.
 */
static void
add_edge(struct adj_node **v_adj, struct adj_node *e_lst, int e, int x)
{
	struct adj_node *edge = NULL;
	
	edge = e_lst + e;
	edge->state = 1;
	if (edge->u == INT_MAX) {
		edge->u = x;
		edge->u_prv = NULL;
		edge->u_nxt = v_adj[x];
		if (v_adj[x] != NULL) {
			if (v_adj[x]->u == x)
				v_adj[x]->u_prv = edge;
			else
				v_adj[x]->v_prv = edge;
		}
		v_adj[x] = edge;
	} else if (edge->v == INT_MAX) {
		edge->v = x;
		edge->v_prv = NULL;
		edge->v_nxt = v_adj[x];
		if (v_adj[x] != NULL) {
			if (v_adj[x]->u == x)
				v_adj[x]->u_prv = edge;
			else
				v_adj[x]->v_prv = edge;
		}
		v_adj[x] = edge;
	} else
		abort();
}

/*
 * Given the adjacency list for each vertex and the list of edges, the
 * edge 'e' is removed from the adjacency list of vertex 'x'.
 */
static void
remove_edge(struct adj_node **v_adj, struct adj_node *e_lst, int e, int x)
{
	struct adj_node *edge = NULL;
	
	edge = e_lst + e;
	edge->state = 0;
	if (edge->u == x) {
		if (edge->u_nxt != NULL) {
			if (edge->u_nxt->u == x)
				edge->u_nxt->u_prv = edge->u_prv;
			else
				edge->u_nxt->v_prv = edge->u_prv;
		}
		if (edge->u_prv != NULL) {
			if (edge->u_prv->u == x)
				edge->u_prv->u_nxt = edge->u_nxt;
			else
				edge->u_prv->v_nxt = edge->u_nxt;
		} else
			v_adj[x] = edge->u_nxt;
		edge->u_nxt = edge->u_prv = NULL;
		edge->u = INT_MAX;
	} else if (edge->v == x) {
		if (edge->v_nxt != NULL) {
			if (edge->v_nxt->u == x)
				edge->v_nxt->u_prv = edge->v_prv;
			else
				edge->v_nxt->v_prv = edge->v_prv;
		}
		if (edge->v_prv != NULL) {
			if (edge->v_prv->u == x)
				edge->v_prv->u_nxt = edge->v_nxt;
			else
				edge->v_prv->v_nxt = edge->v_nxt;
		} else
			v_adj[x] = edge->v_nxt;
		edge->v_nxt = edge->v_prv = NULL;
		edge->v = INT_MAX;
	} else
		abort();
}

double	*A;

/* Comparison function. */
int
_dQcmp(struct adj_node * a, struct adj_node * b)
{
	double aval = a->dQ / sqrt(A[a->u]*A[a->v]);
	double bval = b->dQ / sqrt(A[b->u]*A[b->v]);

	if (aval > bval)
		return 1;

	if (aval < bval)
		return -1;

#ifdef _RAND_CMP
	return a->rand - b->rand;
#else
	return 0;
#endif /* _RAND_CMP */
}
	

/*
 * An implementation with ONE heap and cross-linked adjacency lists. At
 * least 2 times faster when compared with the original implementation.
 */
void
gcfnd(int nv, edge *edge_lst, int ne, int *com_id, int *com_nb, double *cQ)
{
	int u, v, x, e_id, i, *mtmp;
	size_t iter, *mask, *v_dg;
	double Q, maxQ;
	struct adj_node *e, *f, *ex, *e_lst, **v_adj;
	pqueue *pq;

#ifndef _NO_CID
	disjoint_set ds;
#endif /* NO_CID */

#ifdef _PROGRESS_METER
	int mcount;
#endif /* _PROGRESS-METER */

#ifdef _TIMER
	clock_t e_time;
#endif /* _TIMER */

#ifdef _RAND_CMP
	struct timeval t;
	int tmp;
	
	gettimeofday(&t, NULL);
        srand((unsigned) (t.tv_sec ^ t.tv_usec));
#endif /* _RAND_CMP */

	/* Memory allocation: O(V + E) space. */
	/* Adjacency for each vertex, i.e., adjacency list for short. */
	v_adj = (struct adj_node **) malloc(sizeof(struct adj_node *)*nv);
	v_dg = (size_t *) malloc(sizeof(size_t)*nv);
	/* Community find stuff. */
	A = (double *) malloc(sizeof(double)*nv);


	/* Edges allocation. */
	e_lst = (struct adj_node *) malloc(sizeof(struct adj_node)*ne);

		/* Priority queue stuff. */
	pq = pqueue_new(ne, e_lst, _dQcmp);

#ifndef _NO_CID
	/* Disjoint set. */
	ds = make_set(nv);
#endif /* _NO_CID */

	/* Auxiliary stuff. */
	mask = (size_t *) malloc(sizeof(size_t)*nv);
	mtmp = (int *) malloc(sizeof(int)*nv);

	/* Graph loading and initialization: O(V + E) time. */
	for (i = 0; i < nv; i++) {
		v_adj[i] = NULL;
		v_dg[i] = 0;
		A[i] = 0.0;
		
		mask[i] = 0;
	}

	for (i = 0; i < ne; i++) {
		u = edge_lst[i].u;
		v = edge_lst[i].v;

		e = e_lst + i;
		e->id = i;
		e->state = 1;

#ifdef _RAND_CMP
		e->rand = i;
#endif

		e->u = u;
		e->v = v;
		A[u] += 1.0 / ((double) (2 * ne));
		A[v] += 1.0 / ((double) (2 * ne));

		/* Add edge 'i' to the head of the adjacency list of 'u'. */
		e->u_nxt = v_adj[u];
		e->u_prv = NULL;
 		if (v_adj[u] != NULL) {
			if (v_adj[u]->u == u)
				v_adj[u]->u_prv = e;
			else
				v_adj[u]->v_prv = e;
		}
		v_adj[u] = e;
		v_dg[u] ++;

		/* Add edge 'i' to the head of the adjacency list of 'v'. */
		e->v_nxt = v_adj[v];
		e->v_prv = NULL;
		if (v_adj[v] != NULL) {
			if (v_adj[v]->v == v)
				v_adj[v]->v_prv = e;
			else
				v_adj[v]->u_prv = e;
		}
		v_adj[v] = e;
		v_dg[v] ++;
	}

#ifdef _TIMER
	e_time = clock();
#endif /* _TIMER */

	/* 'Q' and 'dQ' initialization: O(V + E) time. */
	maxQ = Q = 0.0;
	for (i = 0; i < nv; i++)
		Q -= A[i]*A[i];

	for (i = 0; i < ne; i++) {
		e = e_lst + i;
		e->dQ = 2 * (1.0 / ((double) (2 * ne))
		    - A[e->u] * A[e->v]);

#ifdef _RAND_CMP
		tmp = rand() % (ne - i) + i;
		e_id = e_lst[tmp].rand;
		e_lst[tmp].rand = e->rand;
		e->rand = e_id;
#endif

		pqueue_push(pq, i);
	}

#ifdef _PROGRESS_METER
	mcount = 0;
#endif /* _PROGRESS_METER */

	/* Find communities. */
	iter = 0;
	while (! pqueue_empty(pq)) {
		iter++;

		/* Extract the edge with maximum 'dQ' from queue. */
		e_id = pqueue_top(pq);
		pqueue_pop(pq);
		e = e_lst + e_id;
		
		/* Check if Q can increase. */
		if (e->dQ < 0.0)
			break; /* Best community structure found. */

		if(e->state == 0) /* A residual edge. */
			continue;

		e->state = 0;

		/* Update modularity 'Q'. */
		Q += e->dQ;
		u = e->u;
		v = e->v;

		if (v_dg[u] > v_dg[v]) {
			x = u;
			u = v;
			v = x;
		}
		
		/* Remove edge from the adjacency lists of 'u' and 'v'. */
		remove_edge(v_adj, e_lst, e_id, u);
		remove_edge(v_adj, e_lst, e_id, v);
		v_dg[u] --;
		v_dg[v] --;

		/* Preprocess 'u' adjacency. */
		for (f = v_adj[u]; f != NULL; f = e_nxt(u,f)) {
			mask[e_adj(u,f)] = iter;
			mtmp[e_adj(u,f)] = f->id;
		}

		/* Add the adjacency of 'u' to the adjacency of 'v'. */
		for (f = v_adj[v]; f != NULL; f = e_nxt(v,f)) {

			if (mask[e_adj(v,f)] == iter) {
				/* 'k' is connected to 'u' and 'v'. */
				f->dQ += e_lst[mtmp[e_adj(v,f)]].dQ;
				x = e_adj(v,f);
				remove_edge(v_adj, e_lst, mtmp[x], x);
				remove_edge(v_adj, e_lst, mtmp[x], u);
				v_dg[x] --;
				v_dg[u] --;
			} else {
				/* 'k' is connected to 'v' but not to 'u'. */
				f->dQ -= 2 * A[u] * A[e_adj(v,f)];
			}

			if (f->dQ < 0)
				pqueue_remove(pq, f->id);
			else
				pqueue_update(pq, f->id);
			mask[e_adj(v,f)] = 0;
		}

		for (f = v_adj[u]; f != NULL;) {
			if(mask[e_adj(u,f)] != iter) {
				f = e_nxt(u,f);
				continue;
			}
			/* Next edge must be known, adjacency will change. */
			ex = e_nxt(u,f);

			/* Move 'e' from 'u' to 'v'. */
			x = e_adj(u,f);
			remove_edge(v_adj, e_lst, f->id, u);
			v_dg[u] --;
			add_edge(v_adj, e_lst, f->id, v);
			v_dg[v] ++;

			/* 'k' is connected to 'u' but not to 'v'. */
			f->dQ -= 2 * A[v] * A[x];

			if (f->dQ < 0)
				pqueue_remove(pq, f->id);
			else
				pqueue_update(pq, f->id);
			f = ex;
		}

		/* Update dendrogram and 'A'. */
#ifndef _NO_CID
		union_set(ds, u, v);
#endif /* _NO_CID */
		A[v] += A[u];

		if (maxQ < Q)
			maxQ = Q;

#ifdef _PROGRESS_METER
		if ((mcount++)%100 == 0)
			fprintf(stderr, "\rcfind: %6.2f%% (%6.5f)",
			    ((double) mcount)/(nv-1) * 100, maxQ);
#endif /* _PROGESS_METER */

	}

#ifdef _PROGRESS_METER
	fprintf(stderr, "\rcfind: 100.00%% (%6.5f)\n", maxQ);
#endif /* _PROGESS_METER */

#ifdef _TIMER
	fprintf(stderr, "%fs seconds\n",
	    (((double) clock()) - ((double) e_time)) / CLOCKS_PER_SEC);
#endif /* _TIMER */

#ifndef _NO_CID
	/* Process communities. */
	*com_nb = 0;
	memset(com_id, 0xFF, sizeof(int)*nv);
	for(i = 0; i < nv; i++){
		x = find_set(ds, i); 
		if(com_id[x] == -1)
			com_id[x] = (*com_nb) ++;
		com_id[i] = com_id[x];
	}
#endif /* _NO_CID */
	*cQ = maxQ;

	/* Memory deallocation. */
	free(v_dg);
	free(v_adj);
	free(e_lst);
	pqueue_free(pq);
#ifndef _NO_CID
	free(ds);
#endif /* _NO_CID */
	free(A);
	free(mask);
	free(mtmp);
}

